/*-
 * Copyright (c) 2006-2007 Daniel Roethlisberger <daniel@roe.ch>
 * Copyright (c) 2000-2004 OMNIKEY GmbH (www.omnikey.com)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

/*
 * OMNIKEY CardMan 4040 a.k.a. CardMan eXtended (cmx) driver.
 *
 * I/O based on Linux driver version 1.1.0 by OMNIKEY GmbH (Dual GPL/BSD).
 * $Omnikey: cm4040_cs.c,v 1.7 2004/10/04 09:08:50 jp Exp $
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>
#include <sys/poll.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
#include <sys/uio.h>
#include <sys/selinfo.h>

#include <sys/module.h>
#include <sys/bus.h>

#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h>

#include <dev/cmx/cmxvar.h>
#include <dev/cmx/cmxreg.h>

#ifdef CMX_DEBUG
#define	DEBUG_printf(dev, fmt, args...) \
	device_printf(dev, "%s: " fmt, __FUNCTION__, ##args)
#else
#define	DEBUG_printf(dev, fmt, args...)
#endif

#define	SPIN_COUNT				1000
#define	WAIT_TICKS				(hz/100)
#define	POLL_TICKS				(hz/10)

/* XXX possibly bogus */
#define	CCID_DRIVER_BULK_DEFAULT_TIMEOUT	(150*hz)
#define	CCID_DRIVER_ASYNC_POWERUP_TIMEOUT	(35*hz)
#define	CCID_DRIVER_MINIMUM_TIMEOUT		(3*hz)

static char	BSRBITS[] = "\020"
	"\001BULK_OUT_FULL"
	"\002BULK_IN_FULL";
static char	POLLBITS[] = "\020"
	"\001POLLIN"
	"\002POLLPRI"
	"\004POLLOUT"
	"\010POLLERR"
	"\020POLLHUP"
	"\040POLLINVAL"
	"\080POLLRDNORM"
	"\160POLLRDBAND"
	"\200POLLWRBAND";

devclass_t cmx_devclass;

static d_open_t		cmx_open;
static d_close_t	cmx_close;
static d_read_t		cmx_read;
static d_write_t	cmx_write;
static d_poll_t		cmx_poll;
#ifdef CMX_INTR
static void		cmx_intr(void *arg); /* XXX check */
#endif

static struct cdevsw cmx_cdevsw = {
	.d_version =	D_VERSION,
	.d_open =	cmx_open,
	.d_close =	cmx_close,
	.d_read =	cmx_read,
	.d_write =	cmx_write,
	.d_poll =	cmx_poll,
	.d_name =	"cmx",
};

/*
 * Initialize the softc structure.  Must be called from
 * the bus specific device allocation routine.
 */
void
cmx_init_softc(device_t dev)
{
	struct cmx_softc *sc = device_get_softc(dev);
	sc->dev = dev;
	sc->timeout = CCID_DRIVER_MINIMUM_TIMEOUT;
}

/*
 * Allocate driver resources.  Must be called from the
 * bus specific device allocation routine.  Caller must
 * ensure to call cmx_release_resources to free the
 * resources when detaching.
 * Return zero if successful, and ENOMEM if the resources
 * could not be allocated.
 */
int
cmx_alloc_resources(device_t dev)
{
	struct cmx_softc *sc = device_get_softc(dev);
#ifdef CMX_INTR
	device_t parent = device_get_parent(dev);
#endif

	sc->ioport = bus_alloc_resource_any(dev, SYS_RES_IOPORT,
	                                    &sc->ioport_rid, RF_ACTIVE);
	if (!sc->ioport) {
		device_printf(dev, "failed to allocate io port\n");
		return ENOMEM;
	}
#ifdef CMX_INTR
	sc->irq = bus_alloc_resource(dev, SYS_RES_IRQ,
		&sc->irq_rid, 0ul, ~0ul, 1, RF_SHAREABLE|RF_ACTIVE);
	if (sc->irq) {
		if (BUS_SETUP_INTR(parent, dev, sc->irq, INTR_TYPE_TTY,
				cmx_intr, sc, &sc->intr_cookie) == 0) {
			/* Do something if successful. */
		} else {
			return ENXIO;
		}
	}
#endif
	sc->bst = rman_get_bustag(sc->ioport);
	sc->bsh = rman_get_bushandle(sc->ioport);

	mtx_init(&sc->mtx, device_get_nameunit(dev),
			"cmx softc lock",
			MTX_DEF | MTX_RECURSE);
	callout_init_mtx(&sc->ch, &sc->mtx, 0);

	return 0;
}

/*
 * Release the resources allocated by cmx_allocate_resources.
 */
void
cmx_release_resources(device_t dev)
{
	struct cmx_softc *sc = device_get_softc(dev);

	mtx_destroy(&sc->mtx);

#ifdef CMX_INTR
	if (sc->irq) {
		bus_deactivate_resource(dev, SYS_RES_IRQ,
				sc->irq_rid, sc->irq);
		bus_release_resource(dev, SYS_RES_IRQ,
				sc->irq_rid, sc->irq);
		sc->irq = NULL;
	}
#endif

	if (sc->ioport) {
		bus_deactivate_resource(dev, SYS_RES_IOPORT,
				sc->ioport_rid, sc->ioport);
		bus_release_resource(dev, SYS_RES_IOPORT,
				sc->ioport_rid, sc->ioport);
		sc->ioport = NULL;
	}
	return;
}

/*
 * Bus independant device attachment routine.  Creates the
 * character device node.
 */
int
cmx_attach(device_t dev)
{
	struct cmx_softc *sc = device_get_softc(dev);

	sc->cdev = make_dev(&cmx_cdevsw, 0, UID_ROOT, GID_WHEEL, 0600,
	                    "cmx%d", device_get_unit(dev));
	if (!sc->cdev) {
		device_printf(dev, "failed to create character device\n");
		return ENOMEM;
	}
	sc->cdev->si_drv1 = sc;

	return 0;
}

/*
 * Bus independant device detachment routine.  Makes sure all
 * allocated resources are freed, callouts disabled and waiting
 * processes unblocked.
 */
int
cmx_detach(device_t dev)
{
	struct cmx_softc *sc = device_get_softc(dev);

	sc->dying = 1;
	callout_drain(&sc->ch);
	wakeup(sc);
	destroy_dev(sc->cdev);

	cmx_release_resources(dev);

	DEBUG_printf(dev, "detached\n");
	return 0;
}

/*
 * Wait for buffer status register events.  If test is non-zero,
 * wait until flags are set, otherwise wait until flags are unset.
 * Will spin SPIN_COUNT times, then sleep until timeout is reached.
 * Returns zero if event happened, EIO if the timeout was reached,
 * and EINTR if the device was detached in the meantime.  When that
 * happens, the caller must quit immediately, since a detach is
 * in progress.
 */
static int
cmx_wait_BSR(struct cmx_softc *sc, uint8_t flags, int test)
{
	int error;

	for (int i = 0; i < SPIN_COUNT; i++) {
		if ((READ_1(sc, REG_OFFSET_BSR) & flags) == (test ? flags : 0))
			return 0;
	}

	for (int i = 0; i * WAIT_TICKS < sc->timeout; i++) {
		if ((READ_1(sc, REG_OFFSET_BSR) & flags) == (test ? flags : 0))
			return 0;
		if ((error = tsleep(sc, PWAIT | PCATCH, "cmxbsr",
						WAIT_TICKS)) != EWOULDBLOCK)
			return error ? error : EINTR;
		if (sc->dying)
			return ENXIO;
	}

	return EIO;
}

#ifdef CMX_INTR
/*
 * Read the sync control register.
 */
static uint8_t
cmx_get_SCR(struct cmx_softc *sc)
{
	return READ_1(sc, REG_OFFSET_SCR);
}
#endif

/*
 * Set the sync control register to val.  Before and after writing
 * to the SCR, we wait for the BSR to not signal BULK_OUT_FULL.
 * Returns zero if successful, or whatever errors cmx_wait_BSR can
 * return.  EINTR signals that the device has been detached in the
 * meantime, and that we should leave the kernel immediately.
 */
static int
cmx_set_SCR(struct cmx_softc *sc, uint8_t val)
{
	int error = 0;

	if ((error = cmx_wait_BSR(sc, BSR_BULK_OUT_FULL, 0)) != 0) {
		return error;
	}

	WRITE_1(sc, REG_OFFSET_SCR, val);

	if ((error = cmx_wait_BSR(sc, BSR_BULK_OUT_FULL, 0)) != 0) {
		return error;
	}

	return 0;
}

/*
 * Periodical callout routine, polling the reader for data
 * availability.  If the reader signals data ready for reading,
 * wakes up the processes which are waiting in select()/poll().
 * Otherwise, reschedules itself with a delay of POLL_TICKS.
 */
static void
cmx_tick(void *xsc)
{
	struct cmx_softc *sc = xsc;
	uint8_t bsr;

	CMX_LOCK(sc);
	if (sc->polling && !sc->dying) {
		bsr = READ_1(sc, REG_OFFSET_BSR);
		DEBUG_printf(sc->dev, "bsr=%b\n", bsr, BSRBITS);
		if ((bsr & BSR_BULK_IN_FULL) == BSR_BULK_IN_FULL) {
			sc->polling = 0;
			selwakeuppri(&sc->sel, PZERO);
			if (sc->read_wait) {
				sc->read_wait = 0;
				wakeup(sc);
			}
		} else {
			callout_reset(&sc->ch, POLL_TICKS, cmx_tick, sc);
		}
	}
	CMX_UNLOCK(sc);
}

/*
 * Open the character device.  Only a single process may open the
 * device at a time.
 */
static int
cmx_open(struct cdev *cdev, int flags, int fmt, struct thread *td)
{
	struct cmx_softc *sc = cdev->si_drv1;

	CMX_LOCK(sc);
	if (sc->open) {
		CMX_UNLOCK(sc);
		return EBUSY;
	}
	sc->open = 1;
	CMX_UNLOCK(sc);

	DEBUG_printf(sc->dev, "opened\n");
	return 0;
}

/*
 * Close the character device.
 */
static int
cmx_close(struct cdev *cdev, int flags, int fmt, struct thread *td)
{
	struct cmx_softc *sc = cdev->si_drv1;

	CMX_LOCK(sc);
	if (!sc->open) {
		DEBUG_printf(sc->dev, "closing closed device!\n");
	}
	sc->open = 0;
	CMX_UNLOCK(sc);

	DEBUG_printf(sc->dev, "closed\n");

	return 0;
}

/*
 * Read from the character device.
 * Returns zero if successful, EDOOFUS if a select()/poll() is
 * already in progress, EINVAL if an attempt was made to read
 * less than CMX_MIN_RDLEN bytes, and EIO if the reader has
 * more bytes available than fit in the internal driver buffer,
 * or any of the errors that cmx_set_SCR can return.
 * Truncates the message from the reader if the userland read
 * buffer is too small.
 */
static int
cmx_read(struct cdev *cdev, struct uio *uio, int flag)
{
	struct cmx_softc *sc = cdev->si_drv1;
	unsigned long bytes_to_read, bytes_read;
	uint8_t uc;
	int error;

	DEBUG_printf(sc->dev, "called (len=%d flag=%d)\n",
		uio->uio_resid, flag);

	CMX_LOCK(sc);
	if (sc->dying) {
		CMX_UNLOCK(sc);
		return ENXIO;
	}
	if (sc->polling) {
		if (flag & O_NONBLOCK) {
			CMX_UNLOCK(sc);
			return (EWOULDBLOCK);
		}
		DEBUG_printf(sc->dev, "sleeping on %p\n", sc);
		sc->read_wait = 1;
		error = msleep(sc, &sc->mtx, PZERO|PCATCH, "cmxrd", 0);
		DEBUG_printf(sc->dev, "woke on %p (error=%d)\n", sc, error);
		sc->read_wait = 0;
		if (sc->dying)
			error = ENXIO;
		if (error) {
			CMX_UNLOCK(sc);
			return error;
		}
	}
	CMX_UNLOCK(sc);

	if (uio->uio_resid == 0) {
		return 0;
	}

	if (uio->uio_resid < CMX_MIN_RDLEN) {
		return EINVAL;
	}

	for (int i = 0; i < 5; i++) {
		if ((error = cmx_wait_BSR(sc, BSR_BULK_IN_FULL, 1)) != 0) {
			return error;
		}
		sc->buf[i] = READ_1(sc, REG_OFFSET_DTR);
		DEBUG_printf(sc->dev, "buf[%02x]=%02x\n", i, sc->buf[i]);
	}

	bytes_to_read = CMX_MIN_RDLEN +
	                (0x000000FF&((char)sc->buf[1])) +
	                (0x0000FF00&((char)sc->buf[2] << 8)) +
	                (0x00FF0000&((char)sc->buf[3] << 16)) +
	                (0xFF000000&((char)sc->buf[4] << 24));
	DEBUG_printf(sc->dev, "bytes_to_read=%lu\n", bytes_to_read);

	if (bytes_to_read > sizeof(sc->buf)) {
		device_printf(sc->dev, "oversized msg from reader (%lu > %d)!\n",
				bytes_to_read, sizeof(sc->buf));
		return EIO;
	}

	bytes_read = MIN(uio->uio_resid, bytes_to_read);
	DEBUG_printf(sc->dev, "bytes_read=%lu\n", bytes_read);

	for (int i = 5; i < bytes_read; i++) {
		if ((error = cmx_wait_BSR(sc, BSR_BULK_IN_FULL, 1)) != 0) {
			return error;
		}
		sc->buf[i] = READ_1(sc, REG_OFFSET_DTR);
		DEBUG_printf(sc->dev, "buf[%02x]=%02x\n", i, sc->buf[i]);
	}

	if ((error = cmx_wait_BSR(sc, BSR_BULK_IN_FULL, 1)) != 0) {
		return error;
	}

	if ((error = cmx_set_SCR(sc, SCR_READER_TO_HOST_DONE)) != 0) {
		return error;
	}

	uc = READ_1(sc, REG_OFFSET_DTR);
	DEBUG_printf(sc->dev, "DTR=%02x\n", uc);

	if ((error = uiomove(sc->buf, bytes_read, uio)) != 0) {
		DEBUG_printf(sc->dev, "uiomove failed\n");
		return error;
	}

	DEBUG_printf(sc->dev, "success\n");

	return 0;
}

/*
 * Write to the character device.
 * Returns zero if successful, EINVAL if more data should be
 * written than fits into the internal driver buffer or less
 * data written than CMX_MIN_WRLEN, or any of the errors that
 * cmx_set_SCR can return.
 */
static int
cmx_write(struct cdev *cdev, struct uio *uio, int flag)
{
	struct cmx_softc *sc = cdev->si_drv1;
	unsigned long bytes_to_write;
	int error;

	DEBUG_printf(sc->dev, "called (len=%d flag=%d)\n", uio->uio_resid, flag);

	if (uio->uio_resid == 0) {
		return 0;
	}

	if (uio->uio_resid < CMX_MIN_WRLEN || uio->uio_resid > sizeof(sc->buf)) {
		return EINVAL;
	}

	bytes_to_write = uio->uio_resid;
	if ((error = uiomove(sc->buf, uio->uio_resid, uio)) != 0) {
		DEBUG_printf(sc->dev, "uiomove failed\n");
		return error;
	}

	DEBUG_printf(sc->dev, "bytes_to_write=%lu cmd=%02x\n", bytes_to_write, sc->buf[0]);

	switch (sc->buf[0]) {
	case CMD_PC_TO_RDR_XFRBLOCK:
	case CMD_PC_TO_RDR_SECURE:
	case CMD_PC_TO_RDR_TEST_SECURE:
	case CMD_PC_TO_RDR_OK_SECURE:
		sc->timeout = CCID_DRIVER_BULK_DEFAULT_TIMEOUT;
		break;

	case CMD_PC_TO_RDR_ICCPOWERON:
		sc->timeout = CCID_DRIVER_ASYNC_POWERUP_TIMEOUT;
		break;

	case CMD_PC_TO_RDR_GETSLOTSTATUS:
	case CMD_PC_TO_RDR_ICCPOWEROFF:
	case CMD_PC_TO_RDR_GETPARAMETERS:
	case CMD_PC_TO_RDR_RESETPARAMETERS:
	case CMD_PC_TO_RDR_SETPARAMETERS:
	case CMD_PC_TO_RDR_ESCAPE:
	case CMD_PC_TO_RDR_ICCCLOCK:
	default:
		sc->timeout = CCID_DRIVER_MINIMUM_TIMEOUT;
		break;
	}
	DEBUG_printf(sc->dev, "timeout=%lu\n", sc->timeout);

	if ((error = cmx_set_SCR(sc, SCR_HOST_TO_READER_START)) != 0) {
		return error;
	}

	for (int i = 0; i < bytes_to_write; i++) {
		if ((error = cmx_wait_BSR(sc, BSR_BULK_OUT_FULL, 0)) != 0) {
			return error;
		}
		WRITE_1(sc, REG_OFFSET_DTR, sc->buf[i]);
		DEBUG_printf(sc->dev, "buf[%02x]=%02x\n", i, sc->buf[i]);
	}

	if ((error = cmx_set_SCR(sc, SCR_HOST_TO_READER_DONE)) != 0) {
		return error;
	}

	DEBUG_printf(sc->dev, "success\n");

	return 0;
}

/*
 * Poll handler.
 */
static int
cmx_poll(struct cdev *cdev, int events, struct thread *td)
{
	struct cmx_softc *sc = cdev->si_drv1;
	uint8_t bsr = 0;
	int revents = 0;

	DEBUG_printf(sc->dev, "called (events=%b)\n", events, POLLBITS);

	revents = events & (POLLOUT | POLLWRNORM);
	if (events & (POLLIN | POLLRDNORM)) {
		bsr = READ_1(sc, REG_OFFSET_BSR);
		if (bsr & BSR_BULK_IN_FULL) {
			revents |= events & (POLLIN | POLLRDNORM);
		} else {
			selrecord(td, &sc->sel);
			CMX_LOCK(sc);
			if (!sc->polling) {
				sc->polling = 1;
				callout_reset(&sc->ch, POLL_TICKS,
						cmx_tick, sc);
			}
			CMX_UNLOCK(sc);
		}
	}

	DEBUG_printf(sc->dev, "success (bsr=%b)\n", bsr, BSRBITS);

	return revents;
}

#ifdef CMX_INTR
/*
 * Interrupt handler.
 */
static void
cmx_intr(void *arg)
{
	struct cmx_softc *sc = (struct cmx_softc *)arg;

	DEBUG_printf(sc->dev, "IRQ: SCR=0x%0x\n", cmx_get_SCR(sc));

	return;
}
#endif

